static gint configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
  URVars update_rect;
  GdkGCValues values;
  PangoFontDescription *pfd;
  static gint first_time_bug = 0;

  first_time_bug++;
  if(first_time_bug<2) return TRUE;

  gtk_widget_get_allocation (widget, &allocation);
  hrow = allocation.height/NUMBER_OF_ROWS;
  wcol = allocation.width/NUMBER_OF_COLUMNS;
  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = allocation.width;
  update_rect.height = allocation.height;
  if(pixmap) g_object_unref (pixmap);
  pixmap = gdk_pixmap_new (gtk_widget_get_window (widget), allocation.width,
			   allocation.height, -1);

  if(gc) g_object_unref (gc);
  gc = gdk_gc_new (pixmap);
  gdk_gc_get_values (gc, &values);
  values.line_width = 2;
  gdk_gc_set_line_attributes (gc,
                              values.line_width,
                              values.line_style,
                              values.cap_style,
                              values.join_style);

  if(layout) g_object_unref (layout);
  layout = gtk_widget_create_pango_layout (widget, NULL);
  pfd = pango_font_description_from_string ("misc fixed 16");
  pango_layout_set_font_description (layout, pfd);
  pango_font_description_free (pfd);

  draw_rectangle (&update_rect, BLUE, SAVEPIX);
  gdk_window_process_updates (gtk_widget_get_window (widget), TRUE);

  update_display (REDRAW);
  return TRUE;
}

static gint expose_event(GtkWidget *widget, GdkEventExpose *event)
{
  gdk_draw_drawable (gtk_widget_get_window (widget),
                     gc,
		     pixmap,
		     event->area.x,
                     event->area.y,
		     event->area.x,
                     event->area.y,
		     event->area.width,
                     event->area.height);
  return TRUE;
}

static void destroy_main_window ()
{
  gtk_widget_destroy (main_window);
}

static void quit_main_window ()
{
  gtk_main_quit ();
}

static void queue_drawn_area(const URVars *urv)
{
  gtk_widget_queue_draw_area (main_drawing_area, urv->x, urv->y,
                              urv->width, urv->height);
}

static void draw_polygon(const APVars *apv, gint ap_color, gint npts)
{
  GdkPoint animation_pts[15];
  gint i;

  for(i=0; i<npts; i++)
   {
    animation_pts[i].x = apv->x[i];
    animation_pts[i].y = apv->y[i];
   }
  new_pen (ap_color);
  gdk_draw_polygon (pixmap, gc, TRUE, animation_pts, npts);
}

static void draw_lines(const APVars *apv, gint ap_color, gint npts)
{
  GdkPoint animation_pts[15];
  gint i;

  for(i=0; i<npts; i++)
   {
    animation_pts[i].x = apv->x[i];
    animation_pts[i].y = apv->y[i];
   }
  new_pen (ap_color);
  gdk_draw_lines (pixmap, gc, animation_pts, npts);
}

int main (int argc, char *argv[])
{
  GtkWidget *window, *drawing_area, *vbox;
  GtkWidget *window_menu_bar, *file_menu_items;
  GtkWidget *root_file_menu, *file_menu;
  GtkWidget *root_help_menu, *help_menu, *help_menu_items;

  system("xmodmap -e 'keysym F10 = F13 F13 F13 F13'");

  gtk_init (&argc, &argv);
  argc_hold = argc;
  argv_hold = argv;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  main_window = window;
  gtk_window_fullscreen (GTK_WINDOW (window));
  g_signal_connect (window, "destroy", G_CALLBACK (quit_kb), NULL);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);
  window_menu_bar = gtk_menu_bar_new ();
  gtk_box_pack_start (GTK_BOX (vbox), window_menu_bar, FALSE, FALSE, 2);
  gtk_widget_show (window_menu_bar);

  root_file_menu = gtk_menu_item_new_with_label ("File");
  gtk_widget_show (root_file_menu);
  gtk_menu_shell_append (GTK_MENU_SHELL (window_menu_bar), root_file_menu);
  file_menu = gtk_menu_new ();
  file_menu_items = gtk_menu_item_new_with_label ("Quit");
  gtk_widget_show (file_menu_items);
  gtk_menu_shell_append (GTK_MENU_SHELL (file_menu), file_menu_items);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (root_file_menu), file_menu);
  g_signal_connect (file_menu_items, "activate",
                    G_CALLBACK (quit_kb), NULL);

  root_help_menu = gtk_menu_item_new_with_label ("Help");
  gtk_widget_show (root_help_menu);
  gtk_menu_shell_append (GTK_MENU_SHELL (window_menu_bar), root_help_menu);
  help_menu = gtk_menu_new ();
  help_menu_items = gtk_menu_item_new_with_label ("KB Help");
  gtk_widget_show (help_menu_items);
  gtk_menu_shell_append (GTK_MENU_SHELL (help_menu), help_menu_items);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (root_help_menu), help_menu);
  g_signal_connect (help_menu_items, "activate",
                    G_CALLBACK (help_kb), NULL);

  drawing_area = gtk_drawing_area_new ();
  main_drawing_area = drawing_area;
  gtk_widget_show (drawing_area);
  gtk_widget_set_size_request (GTK_WIDGET (drawing_area), 200, 200);
  gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);
  gtk_widget_set_events (drawing_area,
                         gtk_widget_get_events (drawing_area)
                         | GDK_BUTTON_PRESS_MASK);
  g_signal_connect (drawing_area, "expose_event",
		    G_CALLBACK (expose_event), NULL);
  g_signal_connect (drawing_area, "configure_event",
		    G_CALLBACK (configure_event), NULL);
  main_mouse_handler_id = g_signal_connect (drawing_area,
                          "button_press_event",
		          G_CALLBACK (bp_event), NULL);
  g_timeout_add (248, timeout_callback, NULL);

  gtk_widget_show (window);
  main_key_handler_id = g_signal_connect (window, "key_press_event",
		        G_CALLBACK (kp_event), NULL);

#ifdef __FreeBSD__
 fd_bells = open("/dev/speaker", O_RDWR);
#else
 fd_bells = open("/dev/kb_bells", O_RDWR);
#endif

  gtk_main ();
  if(fd_bells>0)
   close(fd_bells);

  system("xmodmap -e 'keysym F13 = F10 F10 F10 F10'");

  return 0;
}
